% SPEXPR Sparse sparse affine expression.
%   An object S of type spexpr represents the affine expression
%     S.A() * x(:) + S.b()
%   spexpr supports all usual linear and structural operations. Objects of
%   type spexpr should not be instantiated explicitly, use SPVAR and
%   SPCONST instead.
%
%   See also spvar, spconst

classdef spexpr

%    
% DIFFOP - sparse affine expression class
%
% Copyright (c) 2014 Jan Lellmann (J.Lellmann@damtp.cam.ac.uk) / Cambridge Image Analysis
%
% This file is part of the DIFFOP library. DIFFOP may be used for
% non-commercial purposes, and may be freely modified and distributed
% as long as this copyright notice is retained. Use of DIFFOP in a
% commercial environment requires explicit written permission.
%

    properties(GetAccess = protected, SetAccess = protected)
        A_
        b_       
        sz_
    end
    
    methods
        function obj = spexpr(what,varargin)
            switch what,
                case 'raw',
                    [obj.A_,obj.b_,obj.sz_] = varargin{1:3};
                case 'var',
                    if numel(varargin) > 1
                        error('size must be specified as a vector');
                    end
                    s = normalize_size(varargin{1});
                    obj.A_ = speye(prod(s));
                    obj.b_ = zeros(prod(s),1);
                    obj.sz_ = s;
                case 'const',
                    require_real(varargin{1});
                    obj.sz_ = size(varargin{1});                    
                    obj.b_ = reshape(varargin{1},[],1);
                    obj.A_ = sparse([],[],[],prod(obj.sz_),varargin{2});
                otherwise,
                    error('invalid argument');
            end
        end
        
        function varargout = subsref(obj, s)
            switch(s(1).type)
                case '()',
                    varargout{:} = structural(obj, @subsref, s);
                case '{}',
                    error('{} not implemented');
                case '.',
                    varargout{:} = builtin('subsref', obj, s);
                otherwise,
                    error('Logic error');
            end
        end
        
        function result = A(obj)
            result = builtin('subsref',obj,struct('type','.','subs','A_'));
        end

        function result = b(obj)
            result = builtin('subsref',obj,struct('type','.','subs','b_'));
        end
        
        function result = subsasgn(obj, s, from)
            switch(s(1).type)
                case '()',
                    from = sptoexpr(from, nvar(obj));
                    ind = reshape(1:prod(obj.sz_), obj.sz_);
                    ind = reshape(subsref(ind, s), [], 1);
                    
                    A = obj.A(); A(ind,:) = from.A();
                    b = obj.b(); b(ind) = from.b();
                    
                    result = spexpr('raw', A, b, obj.sz_);
                case '{}',
                    error('{} not implemented');
                case '.',
                    result = builtin('subsasgn', obj, s, from);
                otherwise,
                    error('Logic error');
            end
        end
        
        function result = end(obj, k, n)
            if k < n
                result = obj.sz_(k);
            else
                result = prod(obj.sz_(k:end));
            end
        end
        
        function disp(obj)
            disp(['variables, dimensions = ' num2str(obj.sz_)]);
            %disp(full(obj.A()));
            %disp(obj.b());
        end
        
        function result = size(obj, which)
            if (nargin > 1)
                result = obj.sz_(which);
            else
                result = obj.sz_;
            end
        end
        
        function result = numel(obj)
            result = prod(size(obj));
        end
        
        function result = plus(a, b)      
            n = nvar(a,b);
            [a,b] = spexpand(sptoexpr(a,n),sptoexpr(b,n));
            result = spexpr('raw', a.A + b.A, a.b + b.b, a.sz_);
        end
        
        function result = minus(a, b)
            n = nvar(a,b);
            [a,b] = spexpand(sptoexpr(a,n),sptoexpr(b,n));
            result = spexpr('raw', a.A - b.A, a.b - b.b, a.sz_);
        end
        
        function result = uplus(a)
            result = a;
        end
        
        function result = uminus(a)
            result = spexpr('raw', - a.A, - a.b, a.sz_);
        end
        
        function result = times(a, b)
            if isa(b,'spexpr')
                t = a; a = b; b = t;
            end
            if ~isnumeric(b)
                error('multiplication requires at least one of the arguments to be numeric');
            end
            require_real(b);
            [a,b] = spexpand(a,b); b = reshape(b,[],1);
            result = spexpr('raw', spdiags(b(:),0,numel(b),numel(b)) * a.A, a.b .* b(:), a.sz_);
        end        
        
        function result = rdivide(a, b)
            if ~isnumeric(b)
                error('rdivide requires numeric divisor');
            end
            require_real(b);
            [a,b] = spexpand(a,b); b = reshape(b,[],1);
            result = spexpr('raw', spdiags(1./b(:),0) * a.A, (1./b(:)) .* a.b, a.sz_);
        end
        
        function result = ldivide(a,b)
            result = b ./ a;
        end
        
        function result = mtimes(a, b)
            if numel(size(a)) ~= 2 || numel(size(b)) ~= 2
                error('mtimes requires two-dimensional matrices');
            end
            if isequal(size(a),[1 1]) || isequal(size(b),[1 1])
                result = a .* b; % pass on to element-wise multiplication
                return;
            end
            % the "b" component
            if isa(a,'spexpr')
                require_real(b);
                
                i = a.sz_(1); j = a.sz_(2); k = size(b,2); l = size(a.A,2);
                % transpose: reshapes a.A (ij x l) into (l x ij) [checked]
                % reshape into (li x j)
                % multiply with j x k -> li x k
                % reshape l x ik
                % transpose ik x l
                newA = reshape(reshape(a.A',l*i,j) * b, l, i*k)';
                newb = reshape(reshape(a.b, a.sz_) * b, [], 1);
            elseif isa(b, 'spexpr')
                require_real(a);
                
                i = size(a,1); j = b.sz_(1); k = b.sz_(2); l = size(b.A,2);
                
                newA = reshape(a * reshape(b.A,j,k*l), i*k, l);
                newb = reshape(a * reshape(b.b, b.sz_), [], 1);
            else
                error('Logic error');
            end
                
            result = spexpr('raw',newA,newb,[size(a,1) size(b,2)]);                
        end        
        
        function result = mrdivide(b,a)
            require_real(a);
            result = inv(a) * b;
        end
        
        function result = mldivide(a,b)
            require_real(a);
            result = inv(a) * b;
        end
        
        function result = transpose(a)
            result = structural(a, @transpose);
        end
        
        function result = ctranspose(a)
            % no complex support yet, so pass on to transpose()
            result = transpose(a);
        end
        
        function result = conj(a)
            % everything is real-valued
            result = a;
        end
        
        function result = reshape(obj, varargin)
            result = spexpr('raw', obj.A(), obj.b(), size(reshape(zeros(obj.sz_),varargin{:}))); % slow but covers all special cases
        end
        
        function result = permute(obj, varargin)
            result = structural(obj, @permute, varargin{:});
        end
        
        function result = structural(obj, op, varargin)
            ind = op(reshape(1:prod(obj.sz_),obj.sz_), varargin{:});
            A = obj.A(); b = obj.b();
            result = spexpr('raw', A(reshape(ind,[],1),:), b(reshape(ind,[],1)), size(ind));
        end
        
        function result = squeeze(obj)
            result = obj;
            result.sz_ = normalize_size(result.sz_(result.sz_ ~= 1));
        end
        
        % does not support 'double', 'native' flags yet
        function result = sum(obj, dim)
            if (nargin < 2)
                dim = first_nonsingleton(obj.sz_);
            end
            
            perm = 1:numel(obj.sz_);
            perm = [dim perm(perm ~= dim)];
            objperm = permute(obj,perm); % bring dim to the front
            
            result = ones(1,size(objperm,1)) * objperm; % slow but generic
            
            result.sz_ = [obj.sz_(1:(dim-1)) 1 obj.sz_((dim+1):end)];
        end

        function result = vertcat(varargin)
            result = cat(1, varargin{:});
        end
        
        function result = horzcat(varargin)
            result = cat(2, varargin{:});
        end
        
        function result = cat(dim,varargin)
            ind = {};
            k = 1;
            A = [];
            b = [];
            n = nvar(varargin{:});
            for i = 1:numel(varargin)
                e = sptoexpr(varargin{i},n);
                ind{i} = reshape(k:(k+prod(e.sz_)-1),e.sz_);
                k = k + prod(e.sz_);
                
                A = [A; e.A];
                b = [b; e.b];
            end
            
            ind = cat(dim, ind{:});
            
            result = spexpr('raw',A(reshape(ind,[],1),:),b(reshape(ind,[],1),:),size(ind));
        end
        
        function result = apply(obj,x)
            result = reshape(obj.A() * reshape(x,[],1) + obj.b(), size(obj));
        end
    end    
end

function result = nvar(varargin)
    result = 0;
    for i = 1:numel(varargin)
        if isa(varargin{i},'spexpr')
            result = max(result,size(varargin{i}.A,2));
        end
    end
end

function [a,b] = spexpand(a,b)
    if isequal(size(a),[1 1])
        a = a(ones(size(b)));
    end
    if isequal(size(b),[1 1])
        b = b(ones(size(b)));
    end
end

function a = sptoexpr(a,nvar)
    if isnumeric(a)
        a = spexpr('const',a,nvar);
    end    
end

function require_real(a)
    if ~isnumeric(a) || ~isreal(a) 
        error('real-valued constant required');
    end
end

function sz = normalize_size(sz)
    while numel(sz) < 2
        sz = [sz 1];
    end
end

function result = first_nonsingleton(sz)
    sz = [find(sz ~= 1) 1];
    result = sz(1);
end

% to be implemented:
%  - kron
%  - conv, conv2, filter
